/** \file   bdd_io_dot.c
 *  \author Takahisa Toda
 */
#include <string.h>

#include "my_def.h"
#include "my_hash.h"
#include "bdd_interface.h"
#include "bdd_io_dot.h"

static uintmax_t recdepth = 0; //!< recursion depth

struct linked_node {
  uintptr_t           key;
  struct linked_node  *nx;
};

static void set_zdd_rec(zddp f, uintmax_t *pid, struct linked_node **a, my_hash *h);

/** \brief    Output a ZDD as a dot file, which can be visualized by using Graphviz. 
 *  \param    m   The maximum value of items
 *  \param    f   A non-constant ZDD
 *  \param    out Pointer to an output file
 *  \return   ST_SUCCESS if successful; otherwise, ST_FAILURE.
 */
int zdd_to_dot(uintmax_t m, zddp f, FILE *out)
{
  if(zdd_isconst(f)) {
    ENSURE_TRUE_WARN(false, "constant ZDD node was given");
    return ST_FAILURE;
  }

  const uintmax_t size = zdd_size(f);
  uintmax_t n  = 1; while(n < size) n = n << 1;
  my_hash   *h = ht_create(n);
  ENSURE_TRUE_MSG(h != NULL, "hash creation failed");

  struct linked_node **a = (struct linked_node**)malloc(sizeof(struct linked_node*)*(m+1));
  ENSURE_TRUE_MSG(a != NULL, "memory allocation failed");
  for(uintmax_t i = 0; i <= m; i++) a[i] = NULL;

  uintmax_t id = 0;
  ht_insert((uintptr_t)zdd_bot(), (uintptr_t)(id++), h);
  ht_insert((uintptr_t)zdd_top(), (uintptr_t)(id++), h);
  assert(recdepth == 0);
  set_zdd_rec(f, &id, a, h);
  assert(recdepth == 0);

  fprintf(out, "digraph zdd {\n");
  fprintf(out, "{rank = same; 0 1}\n");
  for(uintmax_t i = 1; i <= m; i++) {
    if(a[i] == NULL) continue;
    fprintf(out, "{rank = same;");
    for(struct linked_node *p = a[i]; p != NULL; p = p->nx) {
      uintmax_t t; ht_search(p->key, (uintptr_t*)&t, h);
      fprintf(out, " %ju", t);
    }
    fprintf(out, "}\n");
  }

  for(uintmax_t i = m; i > 0; i--) {
    for(struct linked_node *p = a[i]; p != NULL;) {
      struct linked_node *t = p->nx; 

      zddp g  = (zddp)p->key; assert(zdd_itemval(g) == i);
      bool res;
      uintmax_t j, j0, j1;

      res = ht_search((uintptr_t)g,         (uintptr_t*)&j,  h);  assert(res == true);
      res = ht_search((uintptr_t)zdd_lo(g), (uintptr_t*)&j0, h);  assert(res == true);
      res = ht_search((uintptr_t)zdd_hi(g), (uintptr_t*)&j1, h);  assert(res == true);

      fprintf(out, "%ju [label = %jd];\n",           j, i);
      fprintf(out, "%ju -> %ju [style = dotted];\n", j, j0);
      fprintf(out, "%ju -> %ju ;\n",                 j, j1);

      free(p);
      p = t;
    } 
  }
  fprintf(out, "1 [shape=box];\n");
  fprintf(out, "0 [shape=box];\n");
  fprintf(out, "}\n");

  ht_destroy(h); free(a);
  return ST_SUCCESS;
}


/** \brief    Set all internal nodes in a ZDD to a and simultaneously assign a unique id to each node.
 *  \param    f     ZDD
 *  \param    pid   Pointer to a unique id
 *  \param    a     Array of linnked lists, in which internal nodes are contained.
 *  \param    h     Pointer to a hash table
 */
static void set_zdd_rec(zddp f, uintmax_t *pid, struct linked_node **a, my_hash *h)
{
 if(ht_search((uintptr_t)f, NULL, h)) return;

  INC_RECDEPTH(recdepth);  
  set_zdd_rec(zdd_lo(f), pid, a, h);
  set_zdd_rec(zdd_hi(f), pid, a, h);
  DEC_RECDEPTH(recdepth);  

  struct linked_node *p = (struct linked_node*)malloc(sizeof(struct linked_node));
  ENSURE_TRUE_MSG(p != NULL, "memory allocation failed");

  itemval v = zdd_itemval(f);
  p->key    = (uintptr_t)f;
  p->nx     = a[v];
  a[v]      = p;

  ENSURE_TRUE_MSG(*pid < UINTPTR_MAX, "integer overflow");
  ht_insert((uintptr_t)f, (uintptr_t)((*pid)++), h);  
}
