/** \file   bdd_io_kf.c
 *  \brief  Implementation of IO functions for data files in Knuth format (for ZDDs)
 *  \author Takahisa Toda
 */
#include <string.h>

#include "my_def.h"
#include "my_hash.h"
#include "bdd_interface.h"
#include "bdd_io_kf.h"

static const int VAL_BASE_NUM   = 10;
static const int INDEX_BASE_NUM = 16;

static inline int cmp_krec      (const void *a, const void*b);
static inline int parse_kf      (char buf[], uintmax_t *pf, itemval *pv, uintmax_t *plo, uintmax_t *phi);
static uintmax_t  zdd_to_kf_rec (zddp f, uintmax_t *pid, my_hash *h, FILE *out);

static uintmax_t recdepth = 0; //!< recursion depth

/**
 * \brief a record for a knuth form
 */
struct krec {
  itemval v;      //!< the label of a node
  uintmax_t idf;  //!< the temporary index of a node
  uintmax_t idlo; //!< the temporary index of the LO child
  uintmax_t idhi; //!< the temporary index of the HI child
  zddp f;         //!< the index of a node, which is assigned by ZDD library.
};

static inline int cmp_krec(const void *a, const void*b)
{
  const struct krec c = *((const struct krec*)a);
  const struct krec d = *((const struct krec*)b);
  
  if      (c.v < d.v) return  1;
  else if (c.v > d.v) return -1;
  else                return  0;
}


/** \brief  Obtain information of a data file, and check consistency of file format.
 *  \param  prow_count    Pointer to the number of rows (equivalently, the number of internal nodes)
 *  \param  pmaxval       Pointer to the maximum value of item values
 *  \param  in            Pointer to a dafa file
 *  \return ST_SUCCESS if successful; otherwise, ST_FAILURE.
 */
int getfileinfo_kf(uintmax_t *prow_count, itemval *pmaxval, FILE *in)
{
  char buf[BUFSIZ]; uintmax_t m = 0; uintmax_t n = 0;
  fseek(in, 0L, SEEK_SET);
  while (fgets(buf, BUFSIZ, in) != NULL) {
    n++;
    // Find the string surrounded by '~' and '?'.
    int i = 0;
    while(buf[i] != '~' && buf[i] != '\0') i++;
    ENSURE_TRUE_MSG(buf[i] == '~', "invalid format\n");
    i++;
    int j = i;
    while(buf[j] != '?' && buf[j] != '\0') j++;
    ENSURE_TRUE_MSG(buf[j] == '?', "invalid format\n");
    buf[j] = '\0';
    uintmax_t v = strtoull(buf+i, NULL, VAL_BASE_NUM);
    ENSURE_TRUE_MSG(v != 0, "invalid format\n");
    buf[j] = '?';
    if(m < v) m = v;
  }

  *prow_count = n; *pmaxval = m;
  return ST_SUCCESS;
}

/** \brief  Parse a row in a Knuth form data file
 *  \param  buf  Array of length BUFSIZ
 *  \param  pf   Pointer to the index of a node
 *  \param  pv   Pointer to an item value associated with a node
 *  \param  plo  Pointer to the index of the LO child
 *  \param  phi  Pointer to the index of the HI child
 *  \return ST_SUCCESS if successful; otherwise, ST_FAILURE.
 */
static inline int parse_kf(char buf[], uintmax_t *pf, itemval *pv, uintmax_t *plo, uintmax_t *phi)
{
  int i = 0; int j = 0;

  // Find a node index
  while(buf[j] != ':' && buf[j] != '\0') j++;
  ENSURE_TRUE_MSG(buf[j] == ':', "invalid format\n");
  buf[j] = '\0';
  *pf = strtoull(buf, NULL, INDEX_BASE_NUM);
  j++;
  i = j;

  // Find the string surrounded by '~' and '?'.
  while(buf[i] != '~' && buf[i] != '\0') i++;
  ENSURE_TRUE_MSG(buf[i] == '~', "invalid format\n");
  i++;
  j = i;
  while(buf[j] != '?' && buf[j] != '\0') j++;
  ENSURE_TRUE_MSG(buf[j] == '?', "invalid format\n");
  buf[j] = '\0';
  *pv = strtoull(buf+i, NULL, VAL_BASE_NUM);
  ENSURE_TRUE_MSG(*pv != 0 && *pv < BDD_MAXITEMVAL, "invalid value\n");
  j++;
  i = j;

  // Find the LO index
  while(buf[j] != ':' && buf[j] != '\0') j++;
  ENSURE_TRUE_MSG(buf[j] == ':', "invalid format\n");
  buf[j] = '\0';
  *plo = strtoull(buf+i, NULL, INDEX_BASE_NUM);
  j++;
  i = j;

  // Find the HI index
  while(buf[j] != ')' && buf[j] != '\0') j++;
  ENSURE_TRUE_MSG(buf[j] == ')', "invalid format\n");
  buf[j] = '\0';
  *phi = strtoull(buf+i, NULL, INDEX_BASE_NUM);
  j++;
  i = j;
 
  return ST_SUCCESS;
}

/** \brief  Construct a ZDD from an input ZDD data in a Knuth format.
 *  \param  row_count   The number of rows (equivalently, the number of internal nodes)
 *  \param  in          Pointer to a dafa file
 *  \return ZDD
 *  \note   row_count must be greater than 0.
 */
zddp zdd_from_kf(uintmax_t row_count, FILE *in)
{ 
  ENSURE_TRUE_MSG(row_count > 0, "there are no internal nodes");
  row_count += 2; // Add the number of terminal nodes. (Hereafter, row_count means the number of all nodes, including terminal nodes.)
  struct krec *a = (struct krec*)malloc(sizeof(struct krec)*row_count);
  ENSURE_TRUE_MSG(a != NULL, "memory allocation failed\n");
  
  uintmax_t n = 1; while(n < row_count) n = n << 1;
  my_hash *h  = ht_create(n);
  ENSURE_TRUE_MSG(h != NULL, "hash creation failed");

  a[0].v    = INFTY;
  a[0].idf  = 0;
  a[0].idlo = 0;
  a[0].idhi = 0;
  a[0].f    = zdd_bot();

  a[1].v    = INFTY;
  a[1].idf  = 1;
  a[1].idlo = 1;
  a[1].idhi = 1;
  a[1].f    = zdd_top();

  // for internal nodes
  char buf[BUFSIZ];
  fseek(in, 0L, SEEK_SET);
  for (uintmax_t i = 2; i < row_count && fgets(buf, BUFSIZ, in) != NULL; i++) {
    parse_kf(buf, &(a[i].idf), &(a[i].v), &(a[i].idlo), &(a[i].idhi));
    a[i].f = BDD_NULL;
  }

  // Sort records in decreasing order with respect to node label
  qsort(a, row_count, sizeof(struct krec), cmp_krec);

  for(uintmax_t i = 0; i < row_count; i++) {ht_insert(a[i].idf, i, h);}
  for(uintmax_t i = 2; i < row_count; i++) {
    uintmax_t pos = 0;
    ht_search(a[i].idlo, &pos, h); zddp lo = a[pos].f;
    ht_search(a[i].idhi, &pos, h); zddp hi = a[pos].f;
    a[i].f = zdd_node(a[i].v, lo, hi);
    ENSURE_TRUE_MSG(a[i].f != BDD_NULL, "ZDD operation failed");
  }

  zddp r = a[row_count-1].f;
  free(a); ht_destroy(h);
  return r;
}

/** \brief  Output a ZDD as a data file in Knuth format.
 *  \param  f      A non-constant ZDD node 
 *  \param  out    Pointer to a dafa file
 *  \return ST_SUCCESS if successful; otherwise, ST_FAILURE.
 */
int zdd_to_kf(zddp f, FILE *out)
{
  ENSURE_TRUE_MSG(!zdd_isconst(f), "constant ZDD node was given");

  const uintmax_t size  = zdd_size(f);
  uintmax_t n   = 1; while(n < size) n = n << 1;
  my_hash   *h  = ht_create(n);
  ENSURE_TRUE_MSG(h != NULL, "hash table creation failed");

  uintmax_t i = 0;
  ht_insert((uintmax_t)zdd_bot(), i++, h);
  ht_insert((uintmax_t)zdd_top(), i++, h);
  assert(recdepth == 0);
  zdd_to_kf_rec(f, &i, h, out);
  assert(recdepth == 0);

  ht_destroy(h); 
  return ST_SUCCESS;
}


static uintmax_t zdd_to_kf_rec(zddp f, uintmax_t *pid, my_hash *h, FILE *out)
{
  uintmax_t t;
  if(ht_search((uintmax_t)f, &t, h)) return t;

  INC_RECDEPTH(recdepth);  
  uintmax_t lo = zdd_to_kf_rec(zdd_lo(f), pid, h, out);
  uintmax_t hi = zdd_to_kf_rec(zdd_hi(f), pid, h, out);
  DEC_RECDEPTH(recdepth);  

  t = (*pid)++;
  fprintf(out, "%ju: (~%jd?%ju:%ju)\n", t, zdd_itemval(f), lo, hi);
  ht_insert((uintmax_t)f, t, h);
  return t;
}
