/** \file   zcomp.c
 *  \brief  Compression of Hypergraphs into ZDDs
 *  \author Takahisa Toda
 *  \see    T.Toda, Fast Compression of Large-scale Hypergraphs for Solving Combinatorial Problems, in Proc. of Sixteenth International Conference on Discovery Science (DS2013), LNAI 8140, pp. 281--293, Singapore (2013).
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "my_def.h"
#include "rng.h"
#include "bdd_io_setfam.h"
#include "bdd_io_kf.h"
#include "bdd_io_dot.h"

#define PROGNAME "ZCOMP"
#define PROGVERS "1.2.1"

#ifdef SELECT_CUDD
DdManager *dd_mgr = NULL; //!< ZDD manager for CUDD
#endif /*SELECT_CUDD*/

#ifdef SELECT_GSL
my_rng *g_rng     = NULL; //!< random number generator for GSL
#endif /*SELECT_GSL*/

#define PRINT_USAGE(p) do{fprintf(stderr, "Usage: %s [option] input-file [output-file]\n", (p)); \
    fprintf(stderr, "\tCompression Method\n");                                                    \
    fprintf(stderr, "\t-z\tzcomp: bottom-up with sorting (default)\n");       \
    fprintf(stderr, "\t-n\tnaive\n");			                                    \
    fprintf(stderr, "\n");                                                    \
    fprintf(stderr, "\tRead Order, available only for naive method\n");       \
    fprintf(stderr, "\t-o\toriginal order (default)\n");       			          \
    fprintf(stderr, "\t-r\trandom order\n");        			                    \
    fprintf(stderr, "\t-f\tforward order (lexicographic order)\n");			      \
    fprintf(stderr, "\t-b\tbackward order\n");          			                \
    fprintf(stderr, "\n");                                                    \
    fprintf(stderr, "\tOutput File Format\n");                                \
    fprintf(stderr, "\t-k\tKnuth file format (default)\n");                   \
    fprintf(stderr, "\t-d\tdot file format\n");                               \
    fprintf(stderr, "\n");                                                    \
    fprintf(stderr, "\tVerification\n");                                      \
    fprintf(stderr, "\t-c\tcomparison with native method\n");                 \
    fprintf(stderr, "\n");                                                    \
  } while(0)

#define Z_METHOD  'z'
#define N_METHOD  'n'

#define O_ORDER   'o'
#define R_ORDER   'r'
#define F_ORDER   'f'
#define B_ORDER   'b'

#define K_FORMAT  'k'
#define D_FORMAT  'd'

#define DO_CHECK  'c'

#if defined(MISC_LOG) || defined(SIZE_LOG)
  char *g_basename; //!< basename of input file
#endif /*defined(MISC_LOG) || defined(SIZE_LOG)*/

int main(int argc, char *argv[])
{   
  int res        = ST_SUCCESS;
  char method    = Z_METHOD;
  char order     = O_ORDER;
  char format    = K_FORMAT;
  bool docheck   = false;
  char *infile   = NULL;
  char *outfile  = NULL;

  /*** INPUT ARGUMENT ***/  
  for(int i = 1; i < argc; i++) {
    if(argv[i][0] == '-') {
      switch (argv[i][1]){
      case Z_METHOD: case N_METHOD:
        method = argv[i][1];
	      break; 
      case O_ORDER: case R_ORDER: case F_ORDER: case B_ORDER:
        order  = argv[i][1];
	      break; 
      case K_FORMAT: case D_FORMAT:
        format = argv[i][1];
	      break; 
      case DO_CHECK:
        docheck  = true;
	      break; 
      case 'v': case '?': case 'h': default:
	      PRINT_USAGE(argv[0]); return ST_SUCCESS;
      }
    } else {
      if     (infile  == NULL)  {infile = argv[i];}
      else if(outfile == NULL)  {outfile = argv[i];}
      else                      {PRINT_USAGE(argv[0]); return  ST_SUCCESS;}
    }
  }
  if(infile == NULL) {PRINT_USAGE(argv[0]); return  ST_SUCCESS;}
#if defined(MISC_LOG) || defined(SIZE_LOG)
  time_t timer        = time(NULL);
  struct tm *execdate = localtime(&timer);
  g_basename          = infile;
  for(int i = 0; infile[i] != '\0'; i++) {if(infile[i] == '/') g_basename = infile+i+1;}
#endif /*defined(MISC_LOG) || defined(SIZE_LOG)*/
#ifdef MISC_LOG
  printf("date\t%s",          asctime(execdate));
  printf("program\t%s-%s\n",  PROGNAME, PROGVERS);
  printf("package\t%s\n",     BDD_PACKAGE);
  printf("input\t%s\n",       infile);
  if(outfile != NULL)         printf("output\t%s\n", outfile);
  if(method == Z_METHOD)      printf("method\tzcomp: bottom-up with sorting\n");
  else {
    printf("method\tnaive\n");
    if(order == O_ORDER)      printf("order\toriginal\n");
    else if(order == R_ORDER) printf("order\trandom\n");
    else if(order == F_ORDER) printf("order\tforward (lex)\n");
    else                      printf("order\tbackward (anti-lex)\n");
  }
  if(format == K_FORMAT)      printf("format\tknuth\n");
  else                        printf("format\tdot\n");
  if(docheck)                 printf("check\tenabled\n");
  printf("\n");
  fflush(stdout);
#endif /*MISC_LOG*/

  FILE *in = fopen(infile, "r");
  ENSURE_TRUE_MSG(in != NULL, "file open failed\n");

  /*** SETUP BDD PACKAGE ***/
  rng_create();
  uintmax_t row_count, entry_count;
  itemval maxval;
  res = getfileinfo(&row_count, &maxval, &entry_count, in);
  ENSURE_SUCCESS(res);
  ENSURE_TRUE_MSG(maxval <= BDD_MAXITEMVAL, "the number of variable too large");
  bdd_init(maxval, 0);

  /*** COMPRESSION OF DATAFILE ***/
  zddp f;
  clock_t startingtime = current_time();
  if(method == Z_METHOD) f = zcomp(row_count, entry_count, in);
  else {
    if     (order == O_ORDER) f = zdd_naive_comp_orig             (row_count, entry_count, in);
    else if(order == R_ORDER) f = zdd_naive_comp_rand             (row_count, entry_count, in);
    else if(order == F_ORDER) f = zdd_naive_comp_with_sorting     (row_count, entry_count, in);
    else                      f = zdd_naive_comp_with_sorting_rev (row_count, entry_count, in);
  }
  ENSURE_TRUE_MSG(f != BDD_NULL, "ZDD construction failed");
  print_elapsed_time(startingtime, "COMP");
#ifdef MISC_LOG
  printf("|zdd|\t%ju\n",      zdd_size(f));
  printf("#sets\t%ju\n",      zdd_card(f));
  printf("\n");
  fflush(stdout);
#endif /*MISC_LOG*/

  /*** VERIFICATION ***/
  if(docheck) {
    zddp g = zdd_naive_comp_orig(row_count, entry_count, in);
    ENSURE_TRUE_MSG(g != BDD_NULL, "ZDD construction failed");
    if(f == g)  printf("check_result\tcorrect\n");
    else        printf("check_result\tNOT_correct\n");
  }

  /*** OUTPUT ZDD ***/  
  if(outfile != NULL) {
    FILE *out = fopen(outfile, "w");
    startingtime = current_time();
    if(format == K_FORMAT)  zdd_to_kf(f, out);
    else                    zdd_to_dot(maxval, f, out);
    print_elapsed_time(startingtime, "OUTPUT");
    fclose(out);
  }

  bdd_quit();
  rng_destroy();
  fclose(in);

  return ST_SUCCESS;
}


